/***************************************************************************
 * Project Name       : FZM
 * File name          : ip_secfg_IWF_VTC.h
 * Module name        : SECfgLib Module
 * Author             : Aswini SG
 * Date created       : 20.03.2013
 * Date modified last :
 * -------------------------------------------------------------------------
 * Description        : SE config lib
 ***************************************************************************
 * Copyright (c) Nokia 2012
 * All rights reserved.
 * Licensed material - property of Nokia
 ***************************************************************************
 * Revision History (newest ontop)
 * Date        | Name of Person | Description
 * ------------+----------------+-------------------------------------------
 **************************************************************************/


#ifndef _ip_secfg_IWF_VTC_h
#define _ip_secfg_IWF_VTC_h
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kthread.h>
#include <linux/init.h>
/* for header struct */
#include <linux/if_ether.h>
#include <linux/ip.h>
#include <linux/udp.h>

#include <linux/in.h>
#include <linux/inet.h>

#define GENLMSG_DATA(glh) ((void *)(NLMSG_DATA(glh) + GENL_HDRLEN))
#define NLA_DATA(na) ((void *)((char*)(na) + NLA_HDRLEN))

#define IPIPV4ADDRESSLENGTH 4
#define IPIPV6ADDRESSLENGTH 16

//SKB Control Buffer(skb->cb[]) to fill Meta data for MBTAC
#define SKB_CB_METADATA_FILLED 38
#define MBTAC_METADATA_FRAG_CNT 39  // Fragment count was filled @ cb[39]
#define MBTAC_METADATA_FRAG_PKT_SIZE_LO 45
#define MBTAC_METADATA_FRAG_PKT_SIZE_HI 46

#ifndef VTC_OK
 #define VTC_OK (0)
 #endif

#ifndef VTC_ERR
 #define VTC_ERR (1)
#endif
#define ERR_DUPLICATE_TEID        11

 #define LTE_MAX_IP_SRC_IDX              4
#define LTE_IP_SRC_IDX_INVALID   255
#define LTM_GTPU_PRT             2152

#define MAX_SGW_IP_ADDRESSES                           64
#define GTPU_TRSW_FAILED                               (MAX_SGW_IP_ADDRESSES + 1)
#define GTPU_TRSW_FREE                                 0
#define GTPU_TRSW_USED                                 1
#define DSCP_MAX                                      64

/*************MBTAC***********/
#define MAX_OPERATORS               2
#define NO_CPLANE_IPS_PER_OPERATOR  2
#define NO_UPLANE_IPS_PER_OPERATOR  2
#define MAX_NO_CPLANE_IP_ADDR       MAX_OPERATORS * NO_CPLANE_IPS_PER_OPERATOR
#define MAX_NO_UPLANE_IP_ADDR       MAX_OPERATORS * NO_UPLANE_IPS_PER_OPERATOR
#define QCI_LEVELS                  26
#define QCI_DEFAULT_LEVELS          5

#define MBTAC_DEFAULT_DSCP1         ((uint8_t)46)
#define MBTAC_DEFAULT_DSCP2         ((uint8_t)26)
#define MBTAC_DEFAULT_DSCP3         ((uint8_t)46)
#define MBTAC_DEFAULT_DSCP4         ((uint8_t)28)
#define MBTAC_DEFAULT_DSCP5         ((uint8_t)34)

// MB-TAC default DSCP configuration 0x400414000000 ... bits 46,34,28,26 set to one
#define MBTAC_DEFAULT_DSCP_CFG      ((uint64_t)0x400414000000)

#define MBTAC_MAX_DSCP              ((uint8_t)64)

#define MBTAC_PKT_ID_ADDR           ((uint8_t)47)
#define MBTAC_PKT_TYPE_ADDR           ((uint8_t)46)
/*************MBTAC***********/

#define VLANIDS_MAX                  ((uint8_t)8)
#define MAX_VLANID	4096

/******L2SWITCH******/

#define MAX_VLAN_FILTERS 16
#define MAX_SIZEOF_PORTNAME 5

/******L2SWITCH******/
extern uint8_t dscp_pcp[DSCP_MAX];
extern uint8_t cp_dscp;
extern uint8_t mp_dscp;
extern uint32_t mp_IpAddr;


static const uint32_t IP_SE_CFG_MAGIC = 0xDEADAFFE;

// k2l: Originally cfg msgs handled by netfp
static const uint32_t IP_SE_COMM_CFG_MAGIC = 0xDEADAACF;

typedef enum
{
     ERR_INVALID_CONFIGURATION                    = 40,
     ERR_INVALID_INDEX_IN_LTE_IP_TABLE,
     ERR_ENTRY_ALREADY_CONFIGURED,
     ERR_TRSW_CONFIG_TABLE_FULL,
     ERR_GTP_MGMT_TRSW_CONFIG_TABLE_FULL,
     ERR_ENTRY_EXIST_IN_GTP_MGMT_TRSW_CONF_TABLE,
     ERR_CONN_EXIST_IN_GTP_MGMT_TRSW_CONF_TABLE,
     ERR_NO_MATCHING_ENTRY,
} sicp_iwf_error_type;


typedef enum {
     SWITCH_SETUP_L2SWITCH_CONFIG = 0x10,
     IP_SETUP_SICP_TEID_SET = 0x90, //definitions for sicp <-> TEID (LTE)
     IP_SETUP_TEID_SICP_SET,
     IP_SETUP_SICP_TEID_GET,
     IP_SETUP_TEID_SICP_GET,
     IP_SETUP_SICP_TEID_DEL,
     IP_SETUP_TEID_SICP_DEL,
     IP_SETUP_TEID_SICP_MOD,
     IP_SETUP_SICP_TEID_MOD,
     IP_SETUP_LTE_UP_IP_SET,
     IP_SETUP_LTE_UP_IP_GET,
     IP_SETUP_LTE_UP_IP_DEL,
     IP_SETUP_LTE_CP_IP_SET,
     IP_SETUP_LTE_CP_IP_GET,
     IP_SETUP_LTE_CP_IP_DEL,
     IP_SETUP_LTE_SP_IP_SET,
     IP_SETUP_LTE_SP_IP_GET,
     IP_SETUP_LTE_SP_IP_DEL,
     IP_SETUP_LTE_MP_IP_SET,
     IP_SETUP_LTE_MP_IP_GET,
     IP_SETUP_LTE_MP_IP_DEL,
     IP_SETUP_LTE_MBTAC_CONFIG,
     IP_SETUP_TEID_SICP_PRINT,
     IP_SETUP_DL_GTP_MGMT_SET,
     IP_SETUP_DL_GTP_MGMT_GET,
     IP_SETUP_DL_GTP_MGMT_DEL,
     IP_SETUP_DL_GTP_MGMT_MOD,
     IP_SETUP_TEID_SICP_CCID_MOD,
     IP_SETUP_CFGTABLE_GTP_MGMT_SET,
     IP_SETUP_CFGTABLE_GTP_MGMT_MOD,
     IP_SETUP_CFGTABLE_GTP_MGMT_GET,
     IP_SETUP_CFGTABLE_GTP_MGMT_DEL,
     IP_SETUP_LTE_PATH_SUPERV_DISABLE,
     IP_SETUP_LTE_IWF_CLEANUP,
     IP_SETUP_LTE_TEST_MODE_SIM,
     IP_SETUP_LTE_TEST_MODE_NOT_SIM,
     IP_SETUP_DSCP_PCP_MAP,
     IP_SETUP_CPLANE_DSCP_PCP_SET,
     IP_SETUP_MPLANE_DSCP_PCP_SET,
     IP_SETUP_LTE_IPSEC_ENABLED,
     IP_SETUP_LTE_IPSEC_DISABLED,
     IP_SETUP_VLANIDS_MAP,
     IP_SETUP_LTE_VTC_NID_SET,
    IP_SETUP_LTE_MULTI_BEARER_IWF_ADD,
    IP_SETUP_LTE_MULTI_BEARER_IWF_MOD,
    IP_SETUP_LTE_MULTI_BEARER_IWF_DEL,
    IP_SETUP_DUMP_NO_OF_CONNECTIONS,
    IP_SETUP_LTE_IWF_ENABLE_PSV,
    IP_SETUP_MCAST_DL_SESSION_ADD,
    IP_SETUP_MCAST_DL_SESSION_DEL,
    IP_SETUP_MCAST_DL_SESSION_RX_ADD,
    IP_SETUP_MCAST_DL_SESSION_RX_DEL,
    IP_SETUP_MCAST_DL_SESSION_DUMP,
#ifdef __PATH_SUPERV_ENABLE__
     IP_SETUP_CFGTABLE_GTP_MGMT_ENABLE,
     IP_SETUP_CFGTABLE_GTP_MGMT_DISABLE,
#endif
    MULTICAST_CFGN_SET,
    MULTICAST_CFGN_GET,
    MULTICAST_CFGN_DEL,
    IWF_HEARTBEAT_REQ
} ip_se_sicp_type;

typedef enum _se_drv_cfg_type_t_
{
     SE_DRV_SE_TO_HOST = 0xF1,
     SE_DRV_CFG_TEID_SICP = 0xFB,         /* LTE specific configuration */
     SE_DRV_CFG_GTPU_MGMT = 0xF3        /* LTE GTPU MGMT specific configuration */
 } se_drv_cfg_type_t;


enum eIpCfgMsgFlags
 {
    ICM_LTE_CFG_DRV_COMPAT = 0x01,  /* Compatibility to old CfgDrv
                                         * (ioctl instead of netlink) */
    ICM_RETURN_NEEDED      = 0x02,  /* Return code needed */
    ICM_RETURN_WITH_DATA   = 0x04   /* returned message contains payload data */
};


int get_cmd_from_index_vtc(int daIndex);

typedef enum
{
   IPv4=4,
   IPv6=6
}IpVersion;

typedef union {
     uint8_t  ipa   [IPIPV4ADDRESSLENGTH];
     uint16_t ipa_16[IPIPV4ADDRESSLENGTH / sizeof(uint16_t)];
     uint32_t ipa_32;
} UIpv4Address_t;

typedef union {
     uint8_t  ipa    [IPIPV6ADDRESSLENGTH];
     uint16_t ipa_16 [IPIPV6ADDRESSLENGTH / sizeof(uint16_t)];
     uint32_t ipa_32 [IPIPV6ADDRESSLENGTH / sizeof(uint32_t)];
     uint64_t ipa_64 [IPIPV6ADDRESSLENGTH / sizeof(uint64_t)];
} UIpv6Address_t;

typedef union {
   UIpv4Address_t ipv4Address;
   UIpv6Address_t ipv6Address;
} UIpAddress_t;

typedef struct
{
   UIpAddress_t addr_u;
   IpVersion version;
   #define ip6_addr32 addr_u.ipv6Address.ipa_32
   #define ip4_addr32 addr_u.ipv4Address.ipa_32
}iwf_ip_addr_t;


/******MBTAC************************/
#ifndef _IWF_IP6_NETINET6_IN6_H_
struct iwf_ip6_in6_addr {
    union {
        uint8_t         __u6_addr8[16];
        uint16_t        __u6_addr16[8];
        uint32_t        __u6_addr32[4];
        uint64_t        __u6_addr64[2];
    } __u6_addr;
};
#endif

typedef struct iwf_ip6_in6_addr iwf_ip6_in6_addr_t;
#ifndef NSN_IP_ADDR_T
#define NSN_IP_ADDR_T
typedef struct
{
    union
    {
        uint32_t           ipv4_addr;
        iwf_ip6_in6_addr_t      ipv6_addr;
    } u;
    uint8_t version;
} nsn_ip_addr_t;
#endif
#define   MBTAC_NH_UDP        17        /**< UDP header */
#define   MBTAC_NH_SCTP       132       /**< SCTP header */
#define   MBTAC_NH_TCP        6         /**< TCP header */
#define   MBTAC_NH_IGMP       2
#define LTM_TOP_PORT_1 319
#define LTM_TOP_PORT_2 320
#define L2_FCS                4
#define IWF_ENET_ETHER_VLAN_TAGGED_HDR_LEN  18
#define IWF_ENET_ETHER_HDR_LEN  14

#define IWF_IPV4_VERSION 4
#define IWF_IPV6_VERSION 6

typedef enum
{
    MBTAC_GET_FP_TMB_MEASUREMENTS = 0,
    MBTAC_SET_DSCP,
    MBTAC_GET_DSCP,
    MBTAC_SET_INCLUDE_L2_OVERHEAD,
    MBTAC_GET_INCLUDE_L2_OVERHEAD,
    MBTAC_GET_DSCP_CONFIGURATION,
}se_mbtac_type;

typedef enum
{
MBTAC_U_PLANE = 0,
MBTAC_C_PLANE ,
MBTAC_S_PLANE ,
MBTAC_MBMS_C_PLANE,
MBTAC_UNKNOWN_TYPE = 255
}mbtac_pkt_type_t;

typedef struct
{

     iwf_ip_addr_t  ip;
    uint32_t      bytesSumUl;
    uint32_t      bytesSumDl;
} mbTacFpTmbDataPerIpAddr_t;

typedef struct
{
    uint32_t                  elapsedTime;
    mbTacFpTmbDataPerIpAddr_t splaneData;
    mbTacFpTmbDataPerIpAddr_t cplaneData[MAX_NO_CPLANE_IP_ADDR];
    mbTacFpTmbDataPerIpAddr_t uplaneData[MAX_NO_UPLANE_IP_ADDR];
    mbTacFpTmbDataPerIpAddr_t mcastcplaneData;
    mbTacFpTmbDataPerIpAddr_t mcastuplaneData;
} mbTacFpTmbData_t;

typedef struct
{
    uint8_t noQciDscpValues;
    uint8_t dscp[QCI_LEVELS];
} mbTacDscp_t;


typedef struct
{
    uint8_t l2_incl_flag;
} mbTacL2Flag_t;

typedef struct
{
    se_mbtac_type type;     // type for MBTAC messages

    union
    {
        mbTacFpTmbData_t        mbTacFpTmbData;
        mbTacDscp_t             mbTacDscp;
        mbTacL2Flag_t           mbTacL2Flag;
        uint64_t                mbTacDscpCfg;
    } u;
} mbtac_api_t;

/******MBTAC*********/

/******L2SWITCH******/
typedef enum
{
    SWITCH_SETUP_VLAN_AWARENESS_ENABLE = 0,
    SWITCH_SETUP_ACCEPTABLE_FRAME_TYPE_SET,
    SWITCH_SETUP_ACCEPTABLE_FRAME_TYPE_GET,
    SWITCH_SETUP_VID_LIST_SET,
    SWITCH_SETUP_VID_LIST_GET
}vlan_filter_type;

typedef enum
{
    VTC_ETHSWITCH_ADMIT_VLAN_TAGGED_ONLY,
    VTC_ETHSWITCH_ADMIT_NOT_VLAN_TAGGED_ONLY,
    VTC_ETHSWITCH_ADMIT_ALL
}ethernet_admit_frame_type;

typedef struct {
    uint16_t low;
    uint16_t high;
} __attribute__((__packed__)) vlan_filter;

typedef struct
{
    bool flag;
    uint8_t no_of_ports;
}switch_vf_params;

typedef struct
{
    uint8_t enable;
} switch_switch_enable;

typedef struct
{
    vlan_filter filter[MAX_VLAN_FILTERS];
    uint8_t port;
    char interface_name[MAX_SIZEOF_PORTNAME];
} switch_vlan_filter;

typedef struct
{
    uint8_t port;
    ethernet_admit_frame_type type;
    char interface_name[MAX_SIZEOF_PORTNAME];
} switch_acceptable_frame_types;

typedef struct
{
    vlan_filter_type type;
    union
    {
        switch_vf_params vlan_aware_enable;
        switch_acceptable_frame_types accept_frame_types;
        switch_vlan_filter vlan_id_list;
    }u;
}eth_switch_vlan_filter;

/******L2SWITCH******/

typedef struct
{
     uint16_t rx_nid;
     uint16_t tx_nid;
     uint16_t rx_cpid;
     uint16_t tx_cpid;
     uint16_t mtype;
} sicp_data_t;

typedef struct
{
     uint32_t teid_dl; //enode-b teid
     uint16_t mtype;
     uint16_t tx_nid;  //SICAD TRS
     uint16_t tx_cpid; //SICAD TRS
     uint16_t rx_nid;  //SICAD FSP
     uint16_t rx_cpid; //SICAD FSP
} ip_iwf_ingress_map;

typedef struct
{
     uint32_t ip_src; /* BTS IP */
     uint32_t ip_dst; /* SGW */
     uint32_t teid_ul;
     uint8_t  dscp;
     uint16_t ccid_ul;
} ip_iwf_egress_map;
typedef struct
{
     iwf_ip_addr_t ip_src; /* BTS IP */
     iwf_ip_addr_t ip_dst; /* SGW */
     uint32_t teid_ul;
     uint8_t  dscp;
     uint16_t ccid_ul;
} vtc_ip_iwf_egress_map;


typedef struct
{
     uint16_t rx_nid;
     uint16_t tx_nid;
     uint16_t rx_cpid;
     uint16_t tx_cpid;
     iwf_ip_addr_t ip_dst;    /* S-GW IP address */
     iwf_ip_addr_t ip_src;    /* BTS IP address  */
} vtc_lte_gtp_mgmt_table_map;


typedef struct
{
    UIpAddress_t addr_u;
    IpVersion version;
#define ip6_addr32 addr_u.ipv6Address.ipa_32
#define ip4_addr32 addr_u.ipv4Address.ipa_32
    uint16_t refCount;
}iwf_ip_addr_t_n;

typedef struct
{
    uint32_t           ip_src_index;
    uint32_t           teid_ul;
    uint16_t           dst_port;      // UDP Dest
    uint16_t           src_port;      // UDP Src
    uint8_t            uip_dst_index;
    uint8_t            uip_src_index;
    uint8_t            ip_ver;
    uint8_t            dscp;
} sicp_ip_data_t;


typedef struct
{
    iwf_ip_addr_t       uip_dst;
    uint8_t            ip_ver;
    uint32_t           teid_ul;
    iwf_ip_addr_t       uip_src;
    uint8_t            dscp;
} sicp_teid_data_t;

typedef struct
{
    iwf_ip_addr_t      uip_src;
    uint8_t            ip_ver;
    uint8_t            index;
} sicp_lte_ip_data_t;

typedef struct _teid_sicp_gtpu_mgmt_data
{
   iwf_ip_addr_t        p_ip;
   uint8_t            ip_ver;
   sicp_data_t        sicp_data;
} teid_sicp_gtpu_mgmt_data_t;

typedef struct
{
  uint16_t           rx_nid;
  uint16_t           tx_nid;
  uint16_t           rx_cpid;
  uint16_t           tx_cpid;
  iwf_ip_addr_t       uip_dst;
  uint32_t           ip_src_index; /*! index is changed to point to ip address */
  iwf_ip_addr_t      uip_src;       /* BTS IP address */
  uint8_t            ip_ver;
} vtc_lte_trsw_config_table_t;

typedef struct k2_gtpu_mgmt_data
{
    iwf_ip_addr_t      p_ip;
    uint8_t            ip_ver;
    sicp_data_t        sicp_data;
} k2_gtpu_mgmt_data_t;

typedef struct
{
    iwf_ip_addr_t      uip_src;
    uint8_t            ip_ver;
    uint8_t            index;
} k2_lte_ip_data_t;


typedef struct
{
    ip_se_sicp_type type;   // type for sicp messages
    union
    {
         uint32_t teid_dl;   // pointer for sicp_tied table, DownLink
         uint32_t p_teid_dl; // pointer for sicp_tied table, DownLink
         uint16_t ccid_ul;   // pointer for sicp_iub_ip and sicp_teid table, UpLink
         uint16_t udp_dest;  // pointer for iub_ip_sicp table, DownLink
    } up;
    union
    {
         sicp_data_t                     teid_sicp_data;
         sicp_teid_data_t                sicp_teid_data;
         sicp_ip_data_t                  sicp_ip_data;
         sicp_lte_ip_data_t              sicp_lte_ip_data;
         k2_lte_ip_data_t                k2_lte_ip_data;
         teid_sicp_gtpu_mgmt_data_t      teid_sicp_gtpu_mgmt_data;
         k2_gtpu_mgmt_data_t             k2_gtpu_mgmt_data;
	 vtc_lte_trsw_config_table_t             trsw_config_table;
         uint64_t                        gtpu_superv_disable;
     } u;
 } ip_sicp_api_t;

typedef struct
{
     uint32_t p_teid_dl; //enode-b teid
     uint32_t p_ip;      //enode-b teid
     uint16_t mtype;
     uint16_t tx_nid;    //SICAD TRS
     uint16_t tx_cpid;   //SICAD TRS
     uint16_t rx_nid;    //SICAD FSP
     uint16_t rx_cpid;   //SICAD FSP
} ip_iwf_gtp_mgmt_error_ind_table_map;


typedef struct _sKernelIpCfgMsgHeader
{
     struct nlmsghdr nlh;
     uint32_t magic;        /* "magic" value for verification */
     uint8_t  msgType;      /* type of the message */
     uint8_t  msgFlags;     /* flags to control the message */
     uint16_t msgSize;      /* message size including this header */
     uint32_t retCode;      /* return code of the configuration command */
} sKernelIpCfgMsgHeader;


typedef struct _sIpSicpApiMsg {
     sKernelIpCfgMsgHeader msgHeader;
     ip_sicp_api_t         ip_sicp_api;
     uint8_t               dscp_pcp[DSCP_MAX];
     uint8_t               cp_dscp;
     uint8_t               mp_dscp;
     uint32_t              mp_IpAddr;
     uint16_t	           vlanMap[VLANIDS_MAX];
     mbtac_api_t           mbTacConfig;
     eth_switch_vlan_filter    vfConfig;
} sIpSicpApiMsg;

//--> k2l

typedef enum
{
    Uplane_1_IP,
    Uplane_2_IP,
} uPlaneType;

/* DL */
typedef union
{
    uint16_t    u16;
    struct
    {
        uint16_t    board_id    : 5;        // Same as in BIP stream ID
        uint16_t    soc_id      : 3;
        uint16_t    reserved    : 8;
    }id;
} device_id_t;
typedef struct
{
    device_id_t device_id;
    uint16_t    local_q_id;
} event_q_id_t;
typedef struct
{
    event_q_id_t    eqid;
    uint16_t        msg_type;
} bip_iwf_cfg_data_t;

typedef struct _nlmsghdr
{
   // Length of message including header
   uint32_t        nlmsg_len;

   // Message type (content type)
   uint16_t        nlmsg_type;

   // Message flags
   uint16_t        nlmsg_flags;

   // Sequence number
   uint32_t        nlmsg_seq;

   // Netlink PID of the proccess sending the message
   uint32_t        nlmsg_pid;
}nlmsghdr1;

typedef struct _sCommonNlHeader
{
    nlmsghdr1 nlh;
    uint32_t magicNum;
    uint16_t seDest;
    uint8_t  qos;
    uint8_t  tagType;
    uint32_t tagVal;
}sCommonNlHeader;

typedef struct _sCommonIpCfgMsgHeader
{
    sCommonNlHeader commonNlHeader;
    uint32_t magic;        /* "magic" value for verification */
    uint8_t  msgType;      /* type of the message */
    uint8_t  msgFlags;     /* flags to control the message */
    uint16_t msgSize;      /* message size including this header */
    uint32_t retCode;      /* return code of the configuration command */
} sCommonIpCfgMsgHeader;

/* Multibearer IWF Set */
typedef struct
{
    uint32_t                dlKey;      // eNBTeid for LTE; UDP dst port for WCDMA
    uint32_t                ccidUl;     // UL TableKey
    uint32_t                peerTeid;   // S-GW TEID (Part of Err Ind Table key)
    uint8_t                 isConnAdd;  // Needed for Err Indication table add
    uint8_t                 ingress_retval;
    uint8_t                 egress_retval;
    uint8_t                 err_ind_retval;
    uint16_t                src_port;
    uint16_t                dst_port;
    bip_iwf_cfg_data_t      teid_bip_data;
    uint8_t                 dscp;
    nsn_ip_addr_t           peerIp;     // S-GW IP   (Part of Err Ind Table key)
    nsn_ip_addr_t           eNBIP;
    uint8_t                 CaOptions;  //LTE1858-FZM
} iwf_set_cfg_api_data_t;

#define MAX_IWF_MULTI_BEARER_SET 20 

typedef struct
{
    ip_se_sicp_type         type;       // type of message

    iwf_set_cfg_api_data_t  iwf_entry[MAX_IWF_MULTI_BEARER_SET];
    uint8_t count;
} iwf_set_cfg_api_t;

/* Multibearer IWF Del */
typedef struct
{
    uint32_t                dlKey;      // eNBTeid for LTE; UDP dst port for WCDMA
    uint32_t                ccidUl;     // UL TableKey
    uint32_t                peerTeid;   // S-GW TEID (Part of Err Ind Table key)
    uint8_t                 ingress_retval;
    uint8_t                 egress_retval;
    uint8_t                 err_ind_retval;
    nsn_ip_addr_t           peerIp;     // S-GW IP   (Part of Err Ind Table key)
} iwf_del_cfg_api_data_t;

#define MAX_IWF_MULTI_BEARER_MOD_DEL 40

typedef struct
{
    ip_se_sicp_type         type;       // type of message

    iwf_del_cfg_api_data_t  iwf_entry[MAX_IWF_MULTI_BEARER_MOD_DEL];
    uint8_t count;
} iwf_del_cfg_api_t;

/* Multibearer IWF Mod */
typedef struct
{
    uint32_t                dlKey;      // eNBTeid for LTE; UDP dst port for WCDMA
    uint32_t                ccidUl;     // UL TableKey
    uint32_t                peerTeid;   // S-GW TEID (Part of Err Ind Table key)
    uint8_t                 isFspMod;   // Needed for Err Indication table mod
    uint8_t                 ingress_retval;
    uint8_t                 egress_retval;
    uint8_t                 err_ind_retval;
    uint8_t                 dscp;
    bip_iwf_cfg_data_t      teid_bip_data;
    nsn_ip_addr_t           peerIp;     // S-GW IP   (Part of Err Ind Table key)
} iwf_mod_cfg_api_data_t;

typedef struct
{
    ip_se_sicp_type         type;       // type of message
    iwf_mod_cfg_api_data_t  iwf_entry[MAX_IWF_MULTI_BEARER_MOD_DEL];
    uint8_t count;
} iwf_mod_cfg_api_t;

typedef struct
{
    uint16_t      rx_nid;
    uint16_t      tx_nid;
    uint16_t      rx_cpid;
    uint16_t      tx_cpid;
    nsn_ip_addr_t ip_dst;       /* S-GW IP address */
    uint32_t      ip_src_index; /*! index is changed to point to ip address */
    nsn_ip_addr_t ip_src;       /* BTS IP address */
} trsw_config_table_t;

typedef struct
{
    ip_se_sicp_type         type;
    trsw_config_table_t trsw_config_table;
} iwf_set_lte_iwfcfg_tbl_t;

typedef struct
{
    ip_se_sicp_type         type;
    uPlaneType              uType;
    nsn_ip_addr_t           uIP;
} iwf_set_cfg_ip_t;

typedef struct
{
    ip_se_sicp_type         type;
    uint8_t		    index;
    nsn_ip_addr_t           uIP;
} iwf_get_cfg_ip_t;

//+++

typedef struct _sIwfSetApiMsg {
    sCommonIpCfgMsgHeader msgHeader;
    iwf_set_cfg_api_t     iwf_cfg_api;
}sIwfSetApiMsg;

typedef struct _sIwfDelApiMsg {
    sCommonIpCfgMsgHeader msgHeader;
    iwf_del_cfg_api_t     iwf_cfg_api;
}sIwfDelApiMsg;

typedef struct _sIwfModApiMsg {
    sCommonIpCfgMsgHeader msgHeader;
    iwf_mod_cfg_api_t     iwf_cfg_api;
}sIwfModApiMsg;

typedef struct _sIwfSetIwfCfgtblMsg{
    sCommonIpCfgMsgHeader msgHeader;
    iwf_set_lte_iwfcfg_tbl_t iwf_cfg_api;
}sIwfSetcfgtblMsg;

typedef struct _sIwfSetIpMsg{
    sCommonIpCfgMsgHeader msgHeader;
    union
    {
	iwf_set_cfg_ip_t     iwf_cfg_api;
	iwf_get_cfg_ip_t     iwf_get_cfg;
    } u;
}sIwfSetIpMsg;
//<--k2l

typedef union uNlinkMsgData
{
    struct
    {
        //cvmx_wqe_t work;
        char       msg[800];
    } send;
    struct
    {
        char msg[800];
    } recv;
} nlinkMsgData;


typedef struct _cd_nlink_data_t_ {
    uint32_t spi;
    uint32_t dstIp;
} cd_nlink_data_t;


void ip_se_sicp_teid_api (sIpSicpApiMsg *msg);

uint8_t ip_se_api(void *msg) ;


static inline sKernelIpCfgMsgHeader* getCfgMsgHeader (void* msg)
{
     sIpSicpApiMsg* temp = (sIpSicpApiMsg*)msg;
     sKernelIpCfgMsgHeader* pMsgHeader = &(temp->msgHeader); //&((sIpSicpApiMsg*)msg->msgHeader);

     return pMsgHeader;
 }


static inline void* getCfgMsgPayload (void* msg)
{
//    sKernelIpCfgMsgHeader* temp = (sKernelIpCfgMsgHeader*)msg;
    sIpSicpApiMsg* temp = (sIpSicpApiMsg*)msg;
//    sKernelIpCfgMsgHeader* pMsgHeader = &(temp->msgHeader); //&((sIpSicpApiMsg*)msg->msgHeader);
      ip_sicp_api_t *payload = &(temp->ip_sicp_api);
//    return msg + sizeof(sKernelIpCfgMsgHeader);
      return payload;
}


/**** Uplink related functions *************/

uint32_t ip_se_sicp_teid_set (sIpSicpApiMsg *msg);
uint32_t ip_se_sicp_teid_mod_dscp (sIpSicpApiMsg *msg);
uint32_t ip_se_sicp_teid_del (sIpSicpApiMsg *msg);
uint32_t ip_se_sicp_teid_get (sIpSicpApiMsg *msg);



/**** Downlink related functions *************/
uint32_t ip_se_teid_sicp_set (sIpSicpApiMsg *msg);
uint32_t ip_se_teid_sicp_mod_ccid (sIpSicpApiMsg *msg);
uint32_t ip_se_teid_sicp_del (sIpSicpApiMsg *msg);
uint32_t ip_se_teid_sicp_get (sIpSicpApiMsg *msg);
uint32_t ip_se_teid_sicp_mod (sIpSicpApiMsg *msg);

/******* Err table ******************/

uint32_t ip_se_gtp_mgmt_error_ind_table_set (sIpSicpApiMsg *msg);
uint32_t ip_se_gtp_mgmt_error_ind_table_get (sIpSicpApiMsg *msg);
uint32_t ip_se_gtp_mgmt_error_ind_table_del (sIpSicpApiMsg *msg);
uint32_t ip_se_gtp_mgmt_error_ind_table_mod (sIpSicpApiMsg *msg);


/***********Trs Sw Config table ***********/
uint32_t ipcs_get_gtp_mgmt_trswConfigTable (sIpSicpApiMsg *msg);
uint32_t ipcs_set_gtp_mgmt_trswConfigTable (sIpSicpApiMsg *msg);
uint32_t ipcs_mod_gtp_mgmt_trswConfigTable (sIpSicpApiMsg *msg);
uint32_t ipcs_del_gtp_mgmt_trswConfigTable (sIpSicpApiMsg *msg);
 #ifdef __PATH_SUPERV_ENABLE__
uint32_t ipcs_gtp_path_mgmt_superv_enable (bool mode);
#endif
/**** IWF lookup functions ***/
//uplink
uint32_t set_sicp_teid_connection (uint16_t ccid_ul, sicp_teid_data_t * sicp_teid_data);
uint32_t del_sicp_teid_connection (uint16_t ccid_ul);
uint32_t mod_sicp_teid_dscp_val (uint16_t ccid_ul, uint8_t dscp);
sicp_ip_data_t * get_sicp_ip_connection (uint16_t ccid_ul, sicp_ip_data_t *sicp_ip_table);
//downlink
uint32_t set_teid_sicp_connection (uint32_t teid_dl, uint16_t rx_nid,  uint16_t tx_nid,
                                   uint16_t rx_cpid, uint16_t tx_cpid, uint16_t mtype);
uint32_t mod_teid_sicp_connection (uint32_t teid_dl, uint16_t rx_nid, uint16_t rx_cpid);
uint32_t del_teid_sicp_connection (uint32_t teid_dl);
uint32_t mod_teid_sicp_connection_ccid(uint32_t teid_dl, uint16_t tx_cpid);
sicp_data_t * get_teid_sicp_connection (uint32_t teid_dl);

uint32_t ipcs_lte_iwf_cleanup (void);
uint32_t set_vtc_nid( sIpSicpApiMsg *msg );
//err table
uint32_t set_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, teid_sicp_gtpu_mgmt_data_t * teid_sicp_gtpu_mgmt_data);
teid_sicp_gtpu_mgmt_data_t * get_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, iwf_ip_addr_t p_ip_dl);
uint32_t del_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, iwf_ip_addr_t p_ip_dl);
uint32_t mod_dl_gtp_mgmt_error_ind_connection (uint32_t p_teid_dl, teid_sicp_gtpu_mgmt_data_t * teid_sicp_gtpu_mgmt_data);

//trsw Config table
uint32_t get_gtp_mgmt_trswConfigTable (uint16_t tx_cpid, iwf_ip_addr_t * ip_dst,
                                               iwf_ip_addr_t *  ip_src, uint8_t * index);

int set_gtp_mgmt_trswConfigTable (uint16_t rx_nid,  uint16_t tx_nid,
                                          uint16_t rx_cpid, uint16_t tx_cpid,
                                                                            iwf_ip_addr_t * ip_dst,iwf_ip_addr_t * ip_src);


int mod_gtp_mgmt_trswConfigTable (uint16_t rx_nid,  uint16_t tx_nid,
                                          uint16_t rx_cpid, uint16_t tx_cpid,
                                                                            iwf_ip_addr_t * ip_dst,iwf_ip_addr_t * ip_src);

int del_gtp_mgmt_trswConfigTable (uint16_t tx_cpid, iwf_ip_addr_t * ip_dst, iwf_ip_addr_t * ip_src);

uint8_t get_teid_0_sicp_connection (iwf_ip_addr_t ip_src, iwf_ip_addr_t ip_dst);

//uplink IP

uint32_t ip_se_lte_ip_get (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_ip_del (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_ip_set (sIpSicpApiMsg *msg);

//C-Plane IP
uint32_t ip_se_lte_cp_ip_get (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_cp_ip_del (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_cp_ip_set (sIpSicpApiMsg *msg);

//S-Plane IP
uint32_t ip_se_lte_sp_ip_get (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_sp_ip_del (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_sp_ip_set (sIpSicpApiMsg *msg);

//M-Plane IP
uint32_t ip_se_lte_mp_ip_get (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_mp_ip_del (sIpSicpApiMsg *msg);
uint32_t ip_se_lte_mp_ip_set (sIpSicpApiMsg *msg);


/*New VTC API -  U-Plane IP*/
int vtc_lte_iwf_sim_mode (bool mode);
int vtc_lte_iwf_ipsec_enable (bool mode);
int vtc_lte_iwf_cleanup(void);

int vtc_lte_up_ip_iwf_add(iwf_ip_addr_t ip_src);
int vtc_lte_up_ip_iwf_get(uint8_t index, iwf_ip_addr_t* ip_address);
int vtc_lte_up_ip_iwf_del(iwf_ip_addr_t ip_address);

/*C-Plane IP*/
int vtc_lte_cp_ip_iwf_add(iwf_ip_addr_t ip_src);
int vtc_lte_cp_ip_iwf_get(uint8_t index, iwf_ip_addr_t* ip_address);
int vtc_lte_cp_ip_iwf_del(iwf_ip_addr_t ip_address);

/* New VTC APIs -Egress*/
int vtc_lte_egress_iwf_add(vtc_ip_iwf_egress_map iwf_egress);
int vtc_lte_egress_iwf_mod(vtc_ip_iwf_egress_map iwf_egress);
int vtc_lte_egress_iwf_del(vtc_ip_iwf_egress_map iwf_egress);
int vtc_lte_egress_iwf_get(vtc_ip_iwf_egress_map* iwf_egress);

/*New VTC API - Ingress*/
int vtc_lte_ingress_iwf_add(ip_iwf_ingress_map iwf_ingress);
int vtc_lte_ingress_iwf_del(ip_iwf_ingress_map iwf_ingress);
int vtc_lte_ingress_iwf_mod(ip_iwf_ingress_map iwf_ingress);
int vtc_lte_ingress_iwf_mod_ccid(ip_iwf_ingress_map iwf_ingress);
int vtc_lte_ingress_iwf_get(sicp_data_t* teid_sicp_data, uint32_t in_teid_dl);

//(TODO: change to vtc names)SW Config table VTC interface API

int ipgw_gtp_mgmt_trswConfigTable_add (vtc_lte_gtp_mgmt_table_map trsw_table);
int ipgw_gtp_mgmt_trswConfigTable_mod (vtc_lte_gtp_mgmt_table_map trsw_table);
int ipgw_gtp_mgmt_trswConfigTable_del (uint16_t tx_cpid,iwf_ip_addr_t ip_dst,iwf_ip_addr_t ip_src);
int ipgw_gtp_mgmt_trswConfigTable_get (vtc_lte_trsw_config_table_t * trsw_config_table);

int ipgw_gtp_mgmt_error_ind_table_add(uint32_t p_teid_dl, uint32_t p_ip, uint16_t rx_nid, uint16_t tx_nid,
                                       uint16_t rx_cpid, uint16_t tx_cpid, uint16_t mtype);
int ipgw_gtp_mgmt_error_ind_table_get(teid_sicp_gtpu_mgmt_data_t* teid_sicp_gtpu_mgmt_data, uint32_t in_p_teid_dl);
int ipgw_gtp_mgmt_error_ind_table_del(ip_iwf_gtp_mgmt_error_ind_table_map iwf_gtp_mgmt_error_ind_table);
int ipgw_gtp_mgmt_error_ind_table_mod(uint32_t p_teid_dl, uint32_t p_ip, uint16_t rx_nid, uint16_t tx_nid,
                                      uint16_t rx_cpid, uint16_t tx_cpid, uint16_t mtype);


/**** DSCP to PCP mapping array ************/
uint32_t ip_se_lte_DscpToPcp(sIpSicpApiMsg *msg);

/**** CplaneDscp Set mapping array ************/
uint32_t ip_se_lte_CplaneDscpToPcp(sIpSicpApiMsg *msg);

/**** MplaneDscp Set mapping array ************/
uint32_t ip_se_lte_MplaneDscpToPcp(sIpSicpApiMsg *msg);


// IWF table dump to proc
uint32_t ip_se_downlink_proc_dump(struct seq_file *m);
uint32_t ip_se_uplink_proc_dump(struct seq_file *m);
uint32_t proc_dump_trswConfigTable(struct seq_file *m);
uint32_t proc_dump_errIndTable(struct seq_file *m);
uint32_t ip_se_lte_up_proc_dump(struct seq_file *m);

/*************MBTAC-API's***********/
uint32_t mbTacInit (void);
uint32_t mbtac_ipse_api (sIpSicpApiMsg *msg,sKernelIpCfgMsgHeader* pMsgHeader );
uint32_t mbTacGetFpTmbMeasurements (sIpSicpApiMsg *swp);
uint32_t mbTacSetDscpValues (sIpSicpApiMsg *swp);
uint32_t mbTacGetDscpValues (sIpSicpApiMsg *swp);
uint32_t mbTacSetFpL2InclFlag (sIpSicpApiMsg *swp);
uint32_t mbTacGetFpL2InclFlag (sIpSicpApiMsg *swp);
uint32_t mbTacGetDscpConfiguration (sIpSicpApiMsg *swp);

/**********MBTAC-PKT-API's*********/
void mbTacCountSplaneBytesDownlink (struct sk_buff *skb);
void mbTacCountCplaneBytesDownlink (struct sk_buff *skb);
void mbTacCountUplaneBytesDownlink (struct sk_buff *skb);

void mbTacMarkingSplaneUplink (struct sk_buff *skb);
void mbTacCountSplaneBytesUplink (struct sk_buff *skb);
void mbTacMarkingCplaneUplink (struct sk_buff *skb);
void mbTacCountCplaneBytesUplink (struct sk_buff *skb);
void mbTacMarkingUplaneUplink (struct sk_buff *skb);
void mbTacCountUplaneBytesUplink (struct sk_buff *skb);
int  mbTacValidateAndUpdateBytes(struct sk_buff *skb,uint16_t fragCnt);

/*************MBTAC***********/

uint32_t ip_se_VlanidToGroupIdMap(sIpSicpApiMsg *msg);
/*****IPSec Pre-Fragmentation Hook*******/
void pre_frag_hook (struct sk_buff *skb);

unsigned short mbTacDlPktLenCalc(struct sk_buff *skb);
unsigned short mbTacUlPktLenCalc(struct sk_buff *skb, uint16_t fragCnt, uint16_t l2HdrLen);

/******L2SWITCH******/

uint32_t l2switch_cfg_api(sIpSicpApiMsg *msg, sKernelIpCfgMsgHeader* pMsgHeader);
uint32_t SetVlanAwarenessFlag(sIpSicpApiMsg *swp);
uint32_t SetAcceptableFrameType(sIpSicpApiMsg *swp);
uint32_t SetVlanIDList(sIpSicpApiMsg *swp);

uint32_t GetVlanRegPerInterface(sIpSicpApiMsg *swp);
/******L2SWITCH******/
#endif
