#include "iwf_table_creation.h" 	
#include <linux/unistd.h>

#define GENLMSG_DATA(glh) ((void *)(NLMSG_DATA(glh) + GENL_HDRLEN))
#define NLA_DATA(na) ((void *)((char*)(na) + NLA_HDRLEN))

//#define __IWF_UPLINK_DEBUG__

shared_hash_table_t  * shared_hash_table = NULL;
sicp_lte_ip_data_t  * sicp_lte_ip = NULL;
sicp_data_t        * teid0_sicp_data = NULL;
sicp_ip_data_t   * lte_sicp_ip_table      = NULL;
iwf_ip_addr_t_n  * ipList = NULL;
logfile_block_names_ptr_t * pSeLogfileBlockNamePointer = NULL;
dl_gtp_mgmt_error_ind_hash_bucket_t * dl_gtp_mgmt_error_ind_table = NULL;
vtc_lte_trsw_config_table_t                 * Strsw_config_table = NULL;
uint8_t                               trsw_config_table_place[MAX_SGW_IP_ADDRESSES] = { GTPU_TRSW_FREE };


uint32_t hash_table_do_hash(uint32_t  key)
{
   uint32_t  hash = 0;
  
   hash = (key % (TEID_SICP_HASH_BUCKETS-1));
   return hash;
}

uint32_t cleanup_teid_sicp_table (void)
{
   teid_sicp_hash_bucket_t *head_bucket;
   teid_sicp_hash_bucket_t *bucket;
   teid_sicp_hash_bucket_t *temp; 
   unsigned int i, j;


    /*  Go through all the buckets and cleanup all existing connections */
    for (i = 0; i < TEID_SICP_HASH_BUCKETS; i++)
    {
       head_bucket = &shared_hash_table[i].teid_sicp_table;
       bucket = head_bucket;
       temp = head_bucket;

       while (bucket)
       {
          for (j = 0; j < TEID_SICP_HASH_ENTRIES_PER_BUCKET; j++)
          {
              teid_sicp_hash_entry_t *entry = &(bucket->entries[j]);
              if (entry->teid != 0)
              {
                 // entry found, delete it
                 entry->teid = 0;
                 entry->sicp_data.rx_nid = 0;
                 entry->sicp_data.tx_nid = 0;
                 entry->sicp_data.rx_cpid = 0;
                 entry->sicp_data.tx_cpid = 0;
                 entry->sicp_data.mtype = 0;
                 bucket->entries_used--;
                 if ((bucket->entries_used == 0) && (bucket != head_bucket))
                 {
                   // bucket is empty, if it's not the first bucket, delete it
                   while (temp->next != bucket)
                           temp = temp->next;
                   temp->next = bucket->next;
                   memset (bucket, 0, sizeof (teid_sicp_hash_bucket_t));
                   kfree (bucket);
		   bucket = temp; 
                 }
             }
         }
         bucket = temp->next;
      } 
   }

    return VTC_OK;
}

uint32_t cleanup_sicp_teid_table (void)
{
     // Clean the table
    memset(lte_sicp_ip_table, 0, sizeof(sicp_ip_data_t) * LTE_WCDMA_CCID_MAX_VALUE);

   return VTC_OK;
}


uint32_t cleanup_dl_gtp_mgmt_error_ind_table (void)
{
    dl_gtp_mgmt_error_ind_hash_bucket_t *head_bucket;
    dl_gtp_mgmt_error_ind_hash_bucket_t *bucket;
    dl_gtp_mgmt_error_ind_hash_bucket_t *temp;
    unsigned int i, j;
 
    for (i = 0; i < TEID_SICP_HASH_BUCKETS; i++)
    {
        head_bucket = &dl_gtp_mgmt_error_ind_table[i];
        bucket = head_bucket;
        temp = head_bucket;

        while (bucket) 
	{
             for (j = 0; j < DL_GTP_MGMT_ERROR_IND_HASH_ENTRIES_PER_BUCKET; j++)
	     {
	         dl_gtp_mgmt_error_ind_hash_entry_t *entry = &(bucket->entries_gtpu_mgmt[j]);
	         if (entry->p_teid)
	         {
	            // entry found, delete it
	            entry->p_teid = 0;
	            entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32   = 0;
	            entry->sicp_gtpu_mgmt_data.p_ip.ip4_addr32   = 0;
	            entry->sicp_gtpu_mgmt_data.sicp_data.rx_nid          = 0;
	            entry->sicp_gtpu_mgmt_data.sicp_data.tx_nid          = 0;
	            entry->sicp_gtpu_mgmt_data.sicp_data.rx_cpid         = 0;
	            entry->sicp_gtpu_mgmt_data.sicp_data.tx_cpid         = 0;
	            entry->sicp_gtpu_mgmt_data.sicp_data.mtype           = 0;
	            bucket->entries_gtpu_mgmt_used--;
	                   
	            if ((bucket->entries_gtpu_mgmt_used == 0) && (bucket != head_bucket))
	            {
	               // bucket is empty, if it's not the first bucket, delete it
	               while (temp->next != bucket)
	                   temp = temp->next;
	               temp->next = bucket->next;
	               memset (bucket, 0, sizeof (dl_gtp_mgmt_error_ind_hash_bucket_t));
	               kfree(bucket);
		       temp = bucket;
	             }
	          }
	       }
	       bucket = temp->next;
	   }
       }
     return VTC_OK;
}  // end of cleanup_**_gtp_mgmt_table



uint32_t ipcs_lte_iwf_cleanup()
{
   uint32_t retVal    = VTC_OK;

   /* Cleanup the ingress table */    
   cleanup_teid_sicp_table(); 

   /* Cleanup the egress table */    
   cleanup_sicp_teid_table(); 

   /* Cleanup the Error Indication table */
   cleanup_dl_gtp_mgmt_error_ind_table();

 return retVal;
}

uint32_t allocate_table_buffer (uint32_t    size, 
                            const char* name, 
                            void**      pTable, 
                            uint32_t*   pLogfileTable)
{
   void * pBuffer;
 
  if (*pTable)
  {   // free memory for re-allocation
      kfree(pTable);
  }
 
  pBuffer = (struct sicp_lte_ip_data_t *) kmalloc (size, GFP_KERNEL);
  if( !pBuffer )
  {
       //printk(" unable to allocate memory to hold mapping table: %s \n",  name);
       return 0;
  }

   //printk (" allocated %u bytes of memory for table :%s \n", size, name);

   memset(pBuffer, 0, size);

   *pTable = pBuffer;

   if (pSeLogfileBlockNamePointer)
   {
      *pLogfileTable = (uint32_t)pBuffer;
   } 
   else
   {
      //printk("WARNING - pSeLogfileBlockNamePointer not initialized: %s \n", name);
   }
  
     return 1;
 }


 int init_lte_iwf_table_dl_ul ()
{
       uint32_t index = 0;

       /* SICP LTE IP table */
       if( !allocate_table_buffer( sizeof(sicp_lte_ip_data_t) * LTE_MAX_IP_SRC_IDX,  
                             LTE_SYSCOM_IWF_IP_DUMP_TABLE_BLOCK_NAME,
                             (void**)&sicp_lte_ip,
                              &(pSeLogfileBlockNamePointer->lte_syscom_iwf_ip_dump_table_ptr) ) )
        {
             return (VTC_ERR);
        }

       /* TRS config table */
        if( !allocate_table_buffer( sizeof(vtc_lte_trsw_config_table_t) * MAX_SGW_IP_ADDRESSES, 
                                          GTPU_MGMT_TRS_CONFIG_TABLE_BLOCK_NAME,
                                        (void**)&Strsw_config_table,
                                       &(pSeLogfileBlockNamePointer->gtpu_mgmt_trs_config_table_ptr) ) )
       {
             return (VTC_ERR);
       }
        
  // Downlink/Ingress mapping for LTM; teid -> sicp
        if( !allocate_table_buffer( sizeof(shared_hash_table_t) * TEID_SICP_HASH_BUCKETS, 
                                  LTE_SYSCOM_IWF_DL_CONNECTION_DUMP_TABLE_BLOCK_NAME,
                                  (void**)&shared_hash_table,
                                   &(pSeLogfileBlockNamePointer->lte_syscom_iwf_dl_connection_dump_table_ptr) ) )
        {
             return (VTC_ERR);
        }
 
         // Uplink/Egress mapping for LTE
	 if( !allocate_table_buffer( sizeof(sicp_ip_data_t) * LTE_WCDMA_CCID_MAX_VALUE, 
	                                     LTE_SYSCOM_IWF_UL_CONNECTION_DUMP_TABLE_BLOCK_NAME,
	                                     (void**)&lte_sicp_ip_table,
	                                     &(pSeLogfileBlockNamePointer->lte_syscom_iwf_ul_connection_dump_table_ptr) ) )
	{
	    return (VTC_ERR);
	}
     // IP table creation
	 if( !allocate_table_buffer( sizeof(iwf_ip_addr_t_n) * LTE_WCDMA_IP_MAX_LIST_VALUE, 
	                                     LTE_SYSCOM_IWF_UL_CONNECTION_DUMP_TABLE_BLOCK_NAME,
	                                     (void**)&ipList,
	                                     NULL) )
	{
	    return (VTC_ERR);
	}

       // Downlink/Ingress mapping for LTM GTP-U management connections; teid -> sicp
       if( !allocate_table_buffer( sizeof(dl_gtp_mgmt_error_ind_hash_bucket_t) * DL_GTP_MGMT_ERROR_IND_HASH_BUCKETS, 
                                   GTPU_MGMT_ERR_IND_TABLE_BLOCK_NAME,
                                  (void**)&dl_gtp_mgmt_error_ind_table,
                                  &(pSeLogfileBlockNamePointer->gtpu_mgmt_err_ind_table_ptr) ) )
       {
           return (VTC_ERR);
       }

       for (index = 0; index < MAX_SGW_IP_ADDRESSES; index++)
      {
         Strsw_config_table[index].rx_nid = 0;
         Strsw_config_table[index].tx_nid = 0;
         Strsw_config_table[index].rx_cpid = 0;
         Strsw_config_table[index].tx_cpid = 0;
         Strsw_config_table[index].uip_dst.ip4_addr32 = 0;
         Strsw_config_table[index].ip_src_index          = LTE_IP_SRC_IDX_INVALID;
         Strsw_config_table[index].uip_src.ip4_addr32 = 0;
       }

       for (index = 0; index < MAX_SGW_IP_ADDRESSES; index++)
       {
	  trsw_config_table_place[index] = GTPU_TRSW_FREE;
       }

      return (0);
}

char * fzm_enet_inet_ntoa_IPv6_r (UIpv6Address_t ina, char * buf)
{
    sprintf(buf, "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x",
            ina.ipa_16[1],
            ina.ipa_16[0],
            ina.ipa_16[3],
            ina.ipa_16[2],
            ina.ipa_16[5],
            ina.ipa_16[4],
            ina.ipa_16[7],
            ina.ipa_16[6]);
    return (buf);
}

char * fzm_enet_inet_ntoa_IPv6_r_linux (struct in6_addr *ina, char * buf)
{
    if ((ina != NULL) && (buf != NULL))
    {
    sprintf(buf, "%04x:%04x:%04x:%04x:%04x:%04x:%04x:%04x",
            ina->s6_addr16[0],
            ina->s6_addr16[1],
            ina->s6_addr16[2],
            ina->s6_addr16[3],
            ina->s6_addr16[4],
            ina->s6_addr16[5],
            ina->s6_addr16[6],
            ina->s6_addr16[7]);
    }
    return (buf);
}
char * fzm_enet_inet_ntoa_r (UIpv4Address_t ina, char *buf)
{
    unsigned char *ucp = (unsigned char *)&ina;

    sprintf(buf, "%d.%d.%d.%d",
            ucp[0] & 0xff,
            ucp[1] & 0xff,
            ucp[2] & 0xff,
            ucp[3] & 0xff);

    return (buf);
}

uint32_t getUserPathIPIndex(iwf_ip_addr_t *trsUpIp)
{
        uint32_t index = 0;
        char     ip_str[40];

        printk("\n%s: trsUpIp = %s\n",__FUNCTION__,((trsUpIp->version == IPv6)? (fzm_enet_inet_ntoa_IPv6_r(trsUpIp->addr_u.ipv6Address, ip_str)) :
                        (fzm_enet_inet_ntoa_r(trsUpIp->addr_u.ipv4Address, ip_str))));

        for (index = 0; index <LTE_MAX_IP_SRC_IDX ; index++)
        {
                printk("\nIP_INDEX_DEBUG:: %s: trsUpIp = %s \n",__FUNCTION__,((trsUpIp->version == IPv6)?
                              (fzm_enet_inet_ntoa_IPv6_r(trsUpIp->addr_u.ipv6Address, ip_str)) :
                              (fzm_enet_inet_ntoa_r(trsUpIp->addr_u.ipv4Address, ip_str))));

                printk("IP_INDEX_DEBUG:: %s: scip_lte_ip[%d].ip_src = %s\n",
                        __FUNCTION__, index,
                         ((sicp_lte_ip[index].uip_src.version == IPv6)?
                         (fzm_enet_inet_ntoa_IPv6_r(sicp_lte_ip[index].uip_src.addr_u.ipv6Address, ip_str)) :
                         (fzm_enet_inet_ntoa_r(sicp_lte_ip[index].uip_src.addr_u.ipv4Address, ip_str))));
                if (ip_is_equal(trsUpIp, &sicp_lte_ip[index].uip_src))
                {
                       printk("IP_INDEX_DEBUG:: %s: index  = %u\n", __FUNCTION__, index);
                       return index;
                }
        }
        printk("IP_INDEX_DEBUG:: %s: index  = %u\n", __FUNCTION__, index);
        return LTE_IP_SRC_IDX_INVALID;
}

//Funciton to check if IPV6 U-plane IP
bool is_ip_U_plane_ipv6(struct in6_addr *ipv6addr)
{
    int i;

#ifdef __IWF_UPLINK_DEBUG__
    char     ip_str[40];
			printk(KERN_INFO "%s: %d Received ip:%s",__FUNCTION__, __LINE__, fzm_enet_inet_ntoa_IPv6_r_linux(ipv6addr, ip_str));
			printk(KERN_INFO "%s: %d Received %08X %08X %08X %08X",__FUNCTION__, __LINE__, 
            ntohl(ipv6addr->s6_addr32[0]),
            ntohl(ipv6addr->s6_addr32[1]),
            ntohl(ipv6addr->s6_addr32[2]),
            ntohl(ipv6addr->s6_addr32[3]));
#endif
    if(!(ipv6addr->s6_addr32[0] | ipv6addr->s6_addr32[1] | ipv6addr->s6_addr32[2] | ipv6addr->s6_addr32[3] ))
    {
       return false;
    }
    for (i = 0; i < LTE_MAX_IP_SRC_IDX; i++)
    {
#ifdef __IWF_UPLINK_DEBUG__
	   printk(KERN_INFO "%s: %d UP plan Ip at index%d :%s",__FUNCTION__, __LINE__, i, 
                fzm_enet_inet_ntoa_IPv6_r(sicp_lte_ip[i].uip_src.addr_u.ipv6Address, ip_str));
	   printk(KERN_INFO "%s: %d Received %08X %08X %08X %08X",__FUNCTION__, __LINE__, 
            (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[0]),
            (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[1]),
            (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[2]),
            (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[3]));
#endif

        if (sicp_lte_ip[i].ip_ver == IPv6)
        {
#ifdef __IWF_UPLINK_DEBUG__
			printk(KERN_INFO "%s: %d",__FUNCTION__, __LINE__);
#endif
            if (    (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[0] == ipv6addr->s6_addr32[0]) &&
                    (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[1] == ipv6addr->s6_addr32[1]) &&
                    (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[2] == ipv6addr->s6_addr32[2]) &&
                    (sicp_lte_ip[i].uip_src.addr_u.ipv6Address.ipa_32[3] == ipv6addr->s6_addr32[3]) )
            {
#ifdef __IWF_UPLINK_DEBUG__
			printk(KERN_INFO "%s: %d",__FUNCTION__, __LINE__);
#endif
                return true;
            }
        }

    }
#ifdef __IWF_UPLINK_DEBUG__
			printk(KERN_INFO "%s: %d",__FUNCTION__, __LINE__);
#endif
    return false;
}

/*****************************************************************************
 * Method Name        : insert_ip
 *
 * Input Parameters   : iwf_ip_addr_t ip
 * 
 * Returns            : 
 * 
 * Description        : Insert the ip into IP table ipList. This table is use to
 *                    : store the egress/ingress src & dst ip                     
 *
*****************************************************************************/
uint8_t insert_ip(iwf_ip_addr_t ip)
{
    uint8_t ret_index = 1; /*index zero should not be used, idex zero is reserved for new entry check*/
    uint8_t first_empty_index = LTE_WCDMA_IP_MAX_LIST_VALUE;

    for(; ret_index < LTE_WCDMA_IP_MAX_LIST_VALUE; ret_index++)
    {
        if(is_ip_addr_null_n(&ipList[ret_index]) == true)
        {
            if(LTE_WCDMA_IP_MAX_LIST_VALUE == first_empty_index)
            {
                first_empty_index = ret_index;
            }
        }
        else
        {
            if(ip_is_equal_n(&ipList[ret_index], &ip) == true)
            {
                ipList[ret_index].refCount++;
                return ret_index;
            }
        }
    }
    if(LTE_WCDMA_IP_MAX_LIST_VALUE != first_empty_index)
    {
        memcpy(&ipList[first_empty_index].addr_u, &ip.addr_u, sizeof(UIpAddress_t));
        ipList[first_empty_index].version = ip.version;
        ret_index = first_empty_index;
        ipList[ret_index].refCount = 1;
    }

    return ret_index;
}

/*****************************************************************************
 * Method Name        : delete_ip
 *
 * Input Parameters   : index
 * 
 * Returns            : 
 * 
 * Description        : Delete the ip from IP table ipList. This table is use to
 *                    : store the egress/ingress src & dst ip                     
 *
 ****************************************************************************/
void delete_ip(uint8_t index)
{
    if(LTE_WCDMA_IP_MAX_LIST_VALUE <= index)
    {
        return;
    }
    if(0 == ipList[index].refCount)
        return;

    ipList[index].refCount--;

    if(0 == ipList[index].refCount)
    {
        memset(&(ipList[index]), 0, sizeof(iwf_ip_addr_t_n));
    }
}


