#include "l2switch_cfg.h"

bool VlanAwareFlag = false;
uint8_t NumofPorts = 2;

svfCfgInfo vlanCfgInfo[MAX_NUM_LOGICAL_INTF];

void initVlanConfig()
{
    uint8_t logicInfIndex = 0, filterIndex = 0;
    for (logicInfIndex = 0; logicInfIndex < MAX_NUM_LOGICAL_INTF; logicInfIndex++)
    {
        vlanCfgInfo[logicInfIndex].AcceptFrameType = ADMIT_ALL;
        for (filterIndex= 0; filterIndex< MAX_VLAN_FILTERS; filterIndex++)
        {
            vlanCfgInfo[logicInfIndex].vidfilter[filterIndex].low = INVALID_VLAN_ID;
            vlanCfgInfo[logicInfIndex].vidfilter[filterIndex].high = INVALID_VLAN_ID;
        }
    }
}

uint32_t l2switch_cfg_api (sIpSicpApiMsg *msg, sKernelIpCfgMsgHeader* pMsgHeader)
{
    uint32_t retCode = ERR_INVALID_CONFIGURATION;
    switch (msg->vfConfig.type)
    {
	case SWITCH_SETUP_VLAN_AWARENESS_ENABLE:
   	    retCode = SetVlanAwarenessFlag(msg);
	    break;
        case SWITCH_SETUP_ACCEPTABLE_FRAME_TYPE_SET:
	    retCode = SetAcceptableFrameType(msg);
	    break;
        case SWITCH_SETUP_VID_LIST_SET:
	    retCode = SetVlanIDList(msg);
	    break;
        case SWITCH_SETUP_VID_LIST_GET:
	    retCode = GetVlanRegPerInterface(msg);
            break;
	default:
            if(0 != printk_ratelimit())
            {
	        printk(KERN_INFO "L2Switch: No Match Found %d", msg->vfConfig.type);
            } 
	    break;
    }
    return retCode;
}

uint32_t SetVlanAwarenessFlag(sIpSicpApiMsg *swp)
{
   uint32_t retVal = VTC_OK;
   VlanAwareFlag = swp->vfConfig.u.vlan_aware_enable.flag;
   NumofPorts = swp->vfConfig.u.vlan_aware_enable.no_of_ports;
   return retVal;
}

uint32_t SetAcceptableFrameType(sIpSicpApiMsg *swp)
{
   uint32_t retVal = VTC_ERR;
   uint8_t portNum = INTF_INVALID;

   portNum = getPortNumPerInterface(swp->vfConfig.u.accept_frame_types.interface_name);
   if(portNum < INTF_INVALID)
   {
	vlanCfgInfo[portNum].AcceptFrameType = swp->vfConfig.u.accept_frame_types.type;
	retVal = VTC_OK;
   }
   return retVal;
}

uint32_t SetVlanIDList(sIpSicpApiMsg *swp)
{
   uint32_t retVal = VTC_ERR;
   uint8_t portNum = INTF_INVALID, filterindex;
   
   portNum = getPortNumPerInterface(swp->vfConfig.u.vlan_id_list.interface_name);
   if(portNum < INTF_INVALID)
   {
	for(filterindex = 0; filterindex < MAX_VLAN_FILTERS; filterindex++)
	{
	    vlanCfgInfo[portNum].vidfilter[filterindex].low = swp->vfConfig.u.vlan_id_list.filter[filterindex].low;
	    vlanCfgInfo[portNum].vidfilter[filterindex].high = swp->vfConfig.u.vlan_id_list.filter[filterindex].high;
	    retVal = VTC_OK;
	}
   } 
   return retVal;
}

uint32_t GetVlanRegPerInterface(sIpSicpApiMsg *swp)
{
   uint32_t retVal = VTC_ERR;
   uint8_t portNum = INTF_INVALID, filterindex;

   portNum = getPortNumPerInterface(swp->vfConfig.u.accept_frame_types.interface_name);
   swp->vfConfig.u.accept_frame_types.type = 0;
   if(portNum < INTF_INVALID)
   {
        swp->vfConfig.u.accept_frame_types.type = vlanCfgInfo[portNum].AcceptFrameType;
        for(filterindex = 0; filterindex < MAX_VLAN_FILTERS; filterindex++)
        {
	    swp->vfConfig.u.vlan_id_list.filter[filterindex].low = vlanCfgInfo[portNum].vidfilter[filterindex].low;
            swp->vfConfig.u.vlan_id_list.filter[filterindex].high = vlanCfgInfo[portNum].vidfilter[filterindex].high;
        }
        retVal = VTC_OK;
   }
   return retVal;
}

uint8_t getPortNumPerInterface(const char* logicalIntfName)
{
   uint8_t retVal = INTF_INVALID;
   if(strncmp(logicalIntfName,"eth0",4) == 0) 
   {
	retVal = INTF_ETH0;
   }
   if(strncmp(logicalIntfName,"eth1",4) == 0) 
   {
	retVal = INTF_ETH1;
   }
   if(strncmp(logicalIntfName,"eth3",4) == 0) 
   {
	retVal = INTF_ETH3;
   }
   return retVal;
}
